__license__ = '''
sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from typing import Literal, TypedDict
import warnings


warnings.warn('importing experimental module')


class WindowEvent(TypedDict, total=False):
    on_afterprstr: str
    on_beforeprstr: str
    on_beforeunload: str
    on_error: str
    on_hashchange: str
    on_load: str
    on_message: str
    on_offline: str
    on_online: str
    on_pagehide: str
    on_pageshow: str
    on_popstate: str
    on_resize: str
    on_storage: str
    on_unhandledrejection: str
    on_unload: str


class FormEvent(TypedDict, total=False):
    on_blur: str
    on_change: str
    on_contextmenu: str
    on_focus: str
    on_input: str
    on_invalid: str
    on_reset: str
    on_search: str
    on_select: str
    on_submit: str


class KeyboardEvent(TypedDict, total=False):
    on_keydown: str
    on_keypress: str
    on_keyup: str


class MouseEvent(TypedDict, total=False):
    on_click: str
    on_dblclick: str
    on_mousedown: str
    on_mousemove: str
    on_mouseout: str
    on_mouseover: str
    on_mouseup: str
    on_wheel: str


class DragEvent(TypedDict, total=False):
    on_drag: str
    on_dragend: str
    on_dragenter: str
    on_dragleave: str
    on_dragover: str
    on_dragstart: str
    on_drop: str


class ClipboardEvent(TypedDict, total=False):
    on_copy: str
    on_cut: str
    on_paste: str


class Event(
    WindowEvent,
    FormEvent,
    KeyboardEvent,
    MouseEvent,
    DragEvent,
    ClipboardEvent,
    total=False,
):
    """Event Attributes for HTML elements."""


class Global(
    Event,
    total=False,
):
    id: str
    accesskey: str
    class_: str
    contenteditable: Literal["true", "false"]
    dir: Literal["ltr", "rtl", "auto"]
    draggable: Literal["true", "false"]
    enterkeyhstr: Literal["enter", "done", "go", "next", "previous", "search", "send"]
    hidden: Literal["true", "false"]
    inert: Literal[""]
    inputmode: Literal["none", "text", "search", "tel", "url", "email", "numeric", "decimal"]
    lang: str
    popover: Literal[""]
    spellcheck: Literal["true", "false"]
    style: str
    tabindex: str
    title: str
    translate: Literal["yes", "no"]


##### NORMAL #####
class A(Global, total=False):
    download: str
    href: str
    hreflang: str
    media: str
    ping: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    rel: Literal[
        "alternate",
        "author",
        "bookmark",
        "external",
        "help",
        "license",
        "next",
        "nofollow",
        "noreferrer",
        "noopener",
        "prev",
        "search",
        "tag",
    ]
    target: str
    type: str


class AUDIO(Global, total=False):
    src: str
    preload: Literal["auto", "metadata", "none"]
    autoplay: Literal[""]
    controls: Literal[""]
    loop: Literal[""]
    muted: Literal[""]
    mediagroup: str


class BLOCKQUOTE(Global, total=False):
    cite: str


class BUTTON(Global, total=False):
    autofocus: Literal[""]
    disabled: Literal[""]
    form: str
    formaction: str
    formenctype: Literal[
        "application/x-www-form-urlencoded",
        "multipart/form-data",
        "text/plain",
    ]
    formmethod: Literal["get", "post"]
    formnovalidate: Literal[""]
    formtarget: str
    popovertarget: str
    popovertargetaction: Literal["show", "hide", "toggle"]
    name: str
    type: Literal["submit", "reset", "button"]
    value: str


class CANVAS(Global, total=False):
    width: str
    height: str


class DATA(Global, total=False):
    value: str


class DEL(Global, total=False):
    cite: str
    datetime: str


class DETAILS(Global, total=False):
    open: Literal[""]


class DIALOG(Global, total=False):
    open: Literal[""]


class FIELDSET(Global, total=False):
    disabled: Literal[""]
    form: str
    name: str


class FORM(Global, total=False):
    accept_charset: str
    action: str
    autocomplete: Literal["on", "off"]
    enctype: Literal[
        "application/x-www-form-urlencoded",
        "multipart/form-data",
        "text/plain",
    ]
    method: Literal["get", "post"]
    name: str
    novalidate: Literal[""]
    rel: Literal[
        "external",
        "help",
        "license",
        "next",
        "nofollow",
        "noopener",
        "noreferrer",
        "opener",
        "prev",
        "search",
    ]
    target: str


class HTML(Global, total=False):
    xmlns: str


class IFRAME(Global, total=False):
    allow: str
    allowfullscreen: Literal[""]
    allowpaymentrequest: Literal[""]
    height: str
    frameborder: str
    loading: Literal["eager", "lazy"]
    name: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    sandbox: Literal[
        "allow-forms",
        "allow-postrer-lock",
        "allow-popups",
        "allow-same-origin",
        "allow-scripts",
        "allow-top-navigation",
    ]
    src: str
    srcdoc: str
    scrolling: str
    width: str


class INS(Global, total=False):
    cite: str
    datetime: str


class LABEL(Global, total=False):
    for_: str
    name: str


class LI(Global, total=False):
    value: str


class MAP(Global, total=False):
    name: str


class METER(Global, total=False):
    form: str
    high: str
    low: str
    max: str
    min: str
    optimum: str
    value: str


class OL(Global, total=False):
    reversed: Literal[""]
    start: str
    type: Literal["1", "a", "A", "i", "I"]


class OPTGROUP(Global, total=False):
    disabled: Literal[""]
    label: str


class OPTION(Global, total=False):
    disabled: Literal[""]
    label: str
    selected: Literal[""]
    value: str


class OUTPUT(Global, total=False):
    for_: str
    form: str
    name: str


class PROGRESS(Global, total=False):
    max: str
    value: str


class Q(Global, total=False):
    cite: str


class SCRIPT(Global, total=False):
    async_: Literal[""]
    crossorigin: Literal["anonymous", "use-credentials"]
    defer: Literal[""]
    stregrity: str
    nomodule: Literal[""]
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    src: str
    type: str


class SELECT(Global, total=False):
    autofocus: Literal[""]
    disabled: Literal[""]
    form: str
    multiple: Literal[""]
    name: str
    required: Literal[""]
    size: str


class STYLE(Global, total=False):
    media: str
    type: Literal["text/css"]


class TD(Global, total=False):
    colspan: str
    headers: str
    rowspan: str


class TEXTAREA(Global, total=False):
    autofocus: Literal[""]
    cols: str
    dirname: str
    disabled: Literal[""]
    form: str
    maxlength: str
    name: str
    placeholder: str
    readonly: Literal[""]
    required: Literal[""]
    rows: str
    wrap: Literal["hard", "soft"]


class TH(Global, total=False):
    abbr: str
    colspan: str
    headers: str
    rowspan: str
    scope: Literal["col", "colgroup", "row", "rowgroup"]


class TIME(Global, total=False):
    datetime: str


class VIDEO(Global, total=False):
    autoplay: Literal[""]
    controls: Literal[""]
    height: str
    loop: Literal[""]
    muted: Literal[""]
    poster: str
    preload: Literal["auto", "metadata", "none"]
    src: str
    width: str


##### VOID #####
class AREA(Global, total=False):
    alt: str
    coords: str
    download: str
    href: str
    ping: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    rel: str
    shape: Literal["rect", "circle", "poly", "default"]
    target: str


class BASE(Global, total=False):
    href: str
    target: str


class COL(Global, total=False):
    span: str


class EMBED(Global, total=False):
    height: str
    src: str
    type: str
    width: str


class IMG(Global, total=False):
    alt: str
    crossorigin: Literal["anonymous", "use-credentials"]
    height: str
    ismap: Literal[""]
    loading: Literal["eager", "lazy"]
    longdesc: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "unsafe-url",
    ]
    sizes: str
    src: str
    srcset: str
    usemap: str
    width: str


class INPUT(Global, total=False):
    accept: str
    alt: str
    autcomplete: Literal["on", "off"]
    autofocus: Literal[""]
    checked: Literal[""]
    dirname: str
    disabled: Literal[""]
    form: str
    formaction: str
    formenctype: Literal[
        "application/x-www-form-urlencoded",
        "multipart/form-data",
        "text/plain",
    ]
    formmethod: Literal["get", "post"]
    formnovalidate: Literal[""]
    formtarget: str
    height: str
    list: str
    max: str
    maxlength: str
    min: str
    minlength: str
    multiple: Literal[""]
    name: str
    pattern: str
    placeholder: str
    popovertarget: str
    popovertargetaction: Literal["show", "hide", "toggle"]
    readonly: Literal[""]
    required: Literal[""]
    size: str
    src: str
    step: str
    type: Literal[
        "checkbox",
        "button",
        "color",
        "date",
        "datetime-local",
        "email",
        "file",
        "hidden",
        "image",
        "month",
        "number",
        "password",
        "radio",
        "range",
        "reset",
        "search",
        "submit",
        "tel",
        "text",
        "time",
        "url",
        "week",
    ]
    value: str
    width: str


class LINK(Global, total=False):
    type: str
    rel: Literal[
        "canonical",
        "alternate",
        "stylesheet",
        "icon",
        "apple-touch-icon",
        "preconnect",
    ]
    crossorigin: Literal[
        "anonymous",
        "use-credentials",
        "",
    ]
    hreflang: str
    href: str
    stregrity: str


class META(Global, total=False):
    name: str
    content: str
    charset: str
    property: str


class PARAM(Global, total=False):
    name: str
    value: str


class SOURCE(Global, total=False):
    media: str
    sizes: str
    src: str
    srcset: str
    type: str


class TRACK(Global, total=False):
    default: Literal[""]
    kind: Literal["subtitles", "captions", "descriptions", "chapters", "metadata"]
    label: str
    src: str
    srclang: str



__license__ = '''
sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
import platform

from sodom import *
from sodom.attrs import Attrs


def __example__():
    with html(lang='en') as doc:
        with head():
            with slot(name='HEADER'):
                meta(charset='utf-8')
                meta(
                    name='viewport',
                    content='width=device-width, initial-scale=1, shrink-to-fit=no',
                )
                slot(name='title')
                link(
                    rel='stylesheet',
                    href='https://cdn.jsdelivr.net/npm/bulma@1.0.0/css/bulma.min.css',
                )
        with body():
            with section() < Attrs(class_='section'):
                with div(class_='container'):
                    title('Hello world!!1!')
                    with p(class_='subtitle'):
                        text('My first website with ')
                        with a(href='https://pypi.org/project/sodom/'):
                            strong('sodom')
                        text(' on your '); strong(f'{platform.node()}/{platform.system()}'); text('!')

    doc.render_html_now()



__license__ = '''

sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from functools import partial

from sodom import literals
from sodom.elements import NormalElement, VoidElement

__all__ = [
    'a',
    'abbr',
    'acronym',
    'address',
    'article',
    'aside',
    'audio',
    'b',
    'bdi',
    'bdo',
    'big',
    'blockquote',
    'body',
    'button',
    'canvas',
    'caption',
    'center',
    'cite',
    'code',
    'colgroup',
    'data',
    'datalist',
    'dd',
    'del_',
    'details',
    'dfn',
    'dialog',
    'dir_',
    'div',
    'dl',
    'dt',
    'em',
    'fieldset',
    'figcaption',
    'figure',
    'font',
    'footer',
    'form',
    'frame',
    'frameset',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'header',
    'hgroup',
    'html',
    'i',
    'iframe',
    'image',
    'ins',
    'kbd',
    'label',
    'legend',
    'li',
    'main',
    'map_',
    'mark',
    'marquee',
    'menu',
    'menuitem',
    'meter',
    'nav',
    'nobr',
    'noembed',
    'noframes',
    'noscript',
    'object_',
    'ol',
    'optgroup',
    'option',
    'output',
    'p',
    'picture',
    'plaintext',
    'portal',
    'pre',
    'progress',
    'q',
    'rb',
    'rp',
    'rt',
    'rtc',
    'ruby',
    's',
    'samp',
    'script',
    'search',
    'section',
    'select',
    'slot',
    'small',
    'span',
    'strike',
    'strong',
    'style',
    'sub',
    'summary',
    'sup',
    'table',
    'tbody',
    'td',
    'template',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'time',
    'title',
    'tr',
    'tt',
    'u',
    'ul',
    'var',
    'video',
    'xmp',

    'area',
    'base',
    'br',
    'col',
    'embed',
    'hr',
    'img',
    'input_',
    'link',
    'meta',
    'param',
    'source',
    'track',
    'wbr',

    'input_datetime_local',
    'input_button',
    'input_checkbox',
    'input_color',
    'input_date',
    'input_email',
    'input_file',
    'input_hidden',
    'input_image',
    'input_month',
    'input_number',
    'input_password',
    'input_radio',
    'input_range',
    'input_reset',
    'input_search',
    'input_submit',
    'input_tel',
    'input_text',
    'input_time',
    'input_url',
    'input_week',
]


##### NORMAL #####
a = partial(NormalElement[literals.A], 'a')
abbr = partial(NormalElement[literals.ABBR], 'abbr')
acronym = partial(NormalElement[literals.ACRONYM], 'acronym')
address = partial(NormalElement[literals.ADDRESS], 'address')
article = partial(NormalElement[literals.ARTICLE], 'article')
aside = partial(NormalElement[literals.ASIDE], 'aside')
audio = partial(NormalElement[literals.AUDIO], 'audio')
b = partial(NormalElement[literals.B], 'b')
bdi = partial(NormalElement[literals.BDI], 'bdi')
bdo = partial(NormalElement[literals.BDO], 'bdo')
big = partial(NormalElement[literals.BIG], 'big')
blockquote = partial(NormalElement[literals.BLOCKQUOTE], 'blockquote')
body = partial(NormalElement[literals.BODY], 'body')
button = partial(NormalElement[literals.BUTTON], 'button')
canvas = partial(NormalElement[literals.CANVAS], 'canvas')
caption = partial(NormalElement[literals.CAPTION], 'caption')
center = partial(NormalElement[literals.CENTER], 'center')
cite = partial(NormalElement[literals.CITE], 'cite')
code = partial(NormalElement[literals.CODE], 'code')
colgroup = partial(NormalElement[literals.COLGROUP], 'colgroup')
data = partial(NormalElement[literals.DATA], 'data')
datalist = partial(NormalElement[literals.DATALIST], 'datalist')
dd = partial(NormalElement[literals.DD], 'dd')
del_ = partial(NormalElement[literals.DEL], 'del')
details = partial(NormalElement[literals.DETAILS], 'details')
dfn = partial(NormalElement[literals.DFN], 'dfn')
dialog = partial(NormalElement[literals.DIALOG], 'dialog')
dir_ = partial(NormalElement[literals.DIR], 'dir')
div = partial(NormalElement[literals.DIV], 'div')
dl = partial(NormalElement[literals.DL], 'dl')
dt = partial(NormalElement[literals.DT], 'dt')
em = partial(NormalElement[literals.EM], 'em')
fieldset = partial(NormalElement[literals.FIELDSET], 'fieldset')
figcaption = partial(NormalElement[literals.FIGCAPTION], 'figcaption')
figure = partial(NormalElement[literals.FIGURE], 'figure')
font = partial(NormalElement[literals.FONT], 'font')
footer = partial(NormalElement[literals.FOOTER], 'footer')
form = partial(NormalElement[literals.FORM], 'form')
frame = partial(NormalElement[literals.FRAME], 'frame')
frameset = partial(NormalElement[literals.FRAMESET], 'frameset')
h1 = partial(NormalElement[literals.H1], 'h1')
h2 = partial(NormalElement[literals.H2], 'h2')
h3 = partial(NormalElement[literals.H3], 'h3')
h4 = partial(NormalElement[literals.H4], 'h4')
h5 = partial(NormalElement[literals.H5], 'h5')
h6 = partial(NormalElement[literals.H6], 'h6')
head = partial(NormalElement[literals.HEAD], 'head')
header = partial(NormalElement[literals.HEADER], 'header')
hgroup = partial(NormalElement[literals.HGROUP], 'hgroup')
html = partial(NormalElement[literals.HTML], 'html')
i = partial(NormalElement[literals.I], 'i')
iframe = partial(NormalElement[literals.IFRAME], 'iframe')
image = partial(NormalElement[literals.IMAGE], 'image')
ins = partial(NormalElement[literals.INS], 'ins')
kbd = partial(NormalElement[literals.KBD], 'kbd')
label = partial(NormalElement[literals.LABEL], 'label')
legend = partial(NormalElement[literals.LEGEND], 'legend')
li = partial(NormalElement[literals.LI], 'li')
main = partial(NormalElement[literals.MAIN], 'main')
map_ = partial(NormalElement[literals.MAP], 'map')
mark = partial(NormalElement[literals.MARK], 'mark')
marquee = partial(NormalElement[literals.MARQUEE], 'marquee')
menu = partial(NormalElement[literals.MENU], 'menu')
menuitem = partial(NormalElement[literals.MENUITEM], 'menuitem')
meter = partial(NormalElement[literals.METER], 'meter')
nav = partial(NormalElement[literals.NAV], 'nav')
nobr = partial(NormalElement[literals.NOBR], 'nobr')
noembed = partial(NormalElement[literals.NOEMBED], 'noembed')
noframes = partial(NormalElement[literals.NOFRAMES], 'noframes')
noscript = partial(NormalElement[literals.NOSCRIPT], 'noscript')
object_ = partial(NormalElement[literals.OBJECT], 'object')
ol = partial(NormalElement[literals.OL], 'ol')
optgroup = partial(NormalElement[literals.OPTGROUP], 'optgroup')
option = partial(NormalElement[literals.OPTION], 'option')
output = partial(NormalElement[literals.OUTPUT], 'output')
p = partial(NormalElement[literals.P], 'p')
picture = partial(NormalElement[literals.PICTURE], 'picture')
plaintext = partial(NormalElement[literals.PLAINTEXT], 'plaintext')
portal = partial(NormalElement[literals.PORTAL], 'portal')
pre = partial(NormalElement[literals.PRE], 'pre')
progress = partial(NormalElement[literals.PROGRESS], 'progress')
q = partial(NormalElement[literals.Q], 'q')
rb = partial(NormalElement[literals.RB], 'rb')
rp = partial(NormalElement[literals.RP], 'rp')
rt = partial(NormalElement[literals.RT], 'rt')
rtc = partial(NormalElement[literals.RTC], 'rtc')
ruby = partial(NormalElement[literals.RUBY], 'ruby')
s = partial(NormalElement[literals.S], 's')
samp = partial(NormalElement[literals.SAMP], 'samp')
script = partial(NormalElement[literals.SCRIPT], 'script')
search = partial(NormalElement[literals.SEARCH], 'search')
section = partial(NormalElement[literals.SECTION], 'section')
select = partial(NormalElement[literals.SELECT], 'select')
slot = partial(NormalElement[literals.SLOT], 'slot')
small = partial(NormalElement[literals.SMALL], 'small')
span = partial(NormalElement[literals.SPAN], 'span')
strike = partial(NormalElement[literals.STRIKE], 'strike')
strong = partial(NormalElement[literals.STRONG], 'strong')
style = partial(NormalElement[literals.STYLE], 'style')
sub = partial(NormalElement[literals.SUB], 'sub')
summary = partial(NormalElement[literals.SUMMARY], 'summary')
sup = partial(NormalElement[literals.SUP], 'sup')
table = partial(NormalElement[literals.TABLE], 'table')
tbody = partial(NormalElement[literals.TBODY], 'tbody')
td = partial(NormalElement[literals.TD], 'td')
template = partial(NormalElement[literals.TEMPLATE], 'template')
textarea = partial(NormalElement[literals.TEXTAREA], 'textarea')
tfoot = partial(NormalElement[literals.TFOOT], 'tfoot')
th = partial(NormalElement[literals.TH], 'th')
thead = partial(NormalElement[literals.THEAD], 'thead')
time = partial(NormalElement[literals.TIME], 'time')
title = partial(NormalElement[literals.TITLE], 'title')
tr = partial(NormalElement[literals.TR], 'tr')
tt = partial(NormalElement[literals.TT], 'tt')
u = partial(NormalElement[literals.U], 'u')
ul = partial(NormalElement[literals.UL], 'ul')
var = partial(NormalElement[literals.VAR], 'var')
video = partial(NormalElement[literals.VIDEO], 'video')
xmp = partial(NormalElement[literals.XMP], 'xmp')


##### VOID #####
area = partial(VoidElement[literals.AREA], 'area')
base = partial(VoidElement[literals.BASE], 'base')
br = partial(VoidElement[literals.BR], 'br')
col = partial(VoidElement[literals.COL], 'col')
embed = partial(VoidElement[literals.EMBED], 'embed')
hr = partial(VoidElement[literals.HR], 'hr')
img = partial(VoidElement[literals.IMG], 'img')
input_ = partial(VoidElement[literals.INPUT], 'input')
link = partial(VoidElement[literals.LINK], 'link')
meta = partial(VoidElement[literals.META], 'meta')
param = partial(VoidElement[literals.PARAM], 'param')
source = partial(VoidElement[literals.SOURCE], 'source')
track = partial(VoidElement[literals.TRACK], 'track')
wbr = partial(VoidElement[literals.WBR], 'wbr')


##### INPUT #####
input_datetime_local = partial(input_, type='datetime-local')
input_button = partial(input_, type='button')
input_checkbox = partial(input_, type='checkbox')
input_color = partial(input_, type='color')
input_date = partial(input_, type='date')
input_email = partial(input_, type='email')
input_file = partial(input_, type='file')
input_hidden = partial(input_, type='hidden')
input_image = partial(input_, type='image')
input_month = partial(input_, type='month')
input_number = partial(input_, type='number')
input_password = partial(input_, type='password')
input_radio = partial(input_, type='radio')
input_range = partial(input_, type='range')
input_reset = partial(input_, type='reset')
input_search = partial(input_, type='search')
input_submit = partial(input_, type='submit')
input_tel = partial(input_, type='tel')
input_text = partial(input_, type='text')
input_time = partial(input_, type='time')
input_url = partial(input_, type='url')
input_week = partial(input_, type='week')



__license__ = '''
sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from abc import ABC, abstractmethod
import html
from contextlib import contextmanager
from contextvars import ContextVar, Token
from functools import partial
from types import TracebackType
from typing import (
    Any,
    Callable,
    ClassVar,
    Generator,
    Iterable,
    Iterator,
    MutableSequence,
    Optional,
    Self,
    Sequence,
)

from sodom.literals import ANY_TAGS, NORMAL_TAGS, VOID_TAGS
from sodom.missing import MISSING


__all__ = [
    '_CURRENT_ELEMENT',
    'isolate',
    'ABCElement',
    'TextWrapper',
    'TagElement',
    'VoidElement',
    'NormalElement',
    'text',
]


_CURRENT_ELEMENT = ContextVar['NormalElement[Any] | None']('_CURRENT_ELEMENT')


@contextmanager
def isolate():
    token = _CURRENT_ELEMENT.set(None)
    try:
        yield
    finally:
        _CURRENT_ELEMENT.reset(token)


class Super:
    __slots__ = (
        '_children',
    )

    _children: Iterable['ABCElement']

    def __init__(self, *children: 'ABCElement') -> None:
        self._children = children

    def __call__(self: Self) -> None:
        for child in self._children:
            child()


class ABCElement(ABC):
    __slots__ = (
        'parent',
        '_ctx',
    )

    REPR_TEMPLATE: ClassVar[Callable[..., str]]

    parent: 'NormalElement | None'
    _ctx: 'ABCRenderContext | None'

    @abstractmethod
    def __eq__(self: Self, other: 'ABCElement | object') -> bool: ...

    def __iter__(self: Self) -> Iterator['ABCElement']:
        yield self

    def __reversed__(self: Self) -> Iterator['NormalElement']:
        yield (current := self)
        while current.parent is not None:
            yield current.parent
            current = current.parent

    @abstractmethod
    def __copy__(self: Self) -> Self: ...

    def __call__(self: Self) -> Self:
        if self.parent is not None:
            self.parent._children.remove(self)
        self.parent = _CURRENT_ELEMENT.get(None)
        if self.parent is not None:
            self.parent._children.append(self)
        return self

    def __mod__(self: Self, ctx: 'ABCRenderContext | None') -> Self:
        self._ctx = ctx
        return self

    def __str__(self: Self) -> str:
        return self.render(ctx=CTX)

    @abstractmethod
    def stream(self: Self, *elems: 'ABCElement', ctx: 'ABCRenderContext') -> Iterator[str]: ...

    def stream_html(
        self: Self,
        *elems: 'ABCElement',
        doctype: str = '<!DOCTYPE html>',
    ) -> Iterator[str]:
        yield from (
            text(doctype, escape=False, use_parent=False)
            .stream(
                self,
                *elems,
                ctx=CTX,
            )
        )

    def render(
        self: Self,
        *elems: 'ABCElement',
        ctx: 'ABCRenderContext',
    ) -> str:
        it = self.stream(
            *elems,
            ctx=ctx,
        )
        return ''.join(it)

    def render_html(
        self: Self,
        *elems: 'ABCElement',
        doctype: str = '<!DOCTYPE html>',
    ) -> str:
        it = self.stream_html(
            *elems,
            doctype=doctype,
        )
        return ''.join(it)

    def render_now(
        self: Self,
        *elems: 'ABCElement',
        ctx: 'ABCRenderContext' = MISSING,
    ) -> None:
        if self._ctx is not None:
            ctx = self._ctx
        if ctx is MISSING:
            raise RuntimeError('ctx is missing')

        import tempfile
        import webbrowser

        tmp_file_path = tempfile.mktemp(
            prefix=(
                f'{ABCElement.render_now.__module__}'
                f'.{ABCElement.render_now.__name__}'
                f'.'
            ),
            suffix='.html',
        )

        with open(tmp_file_path, 'w') as f:
            partitions = self.stream(
                *elems,
                ctx=ctx,
            )
            for partition in partitions:
                f.write(partition)
        webbrowser.open_new_tab(tmp_file_path)

    def render_html_now(
        self: Self,
        *elems: 'ABCElement',
    ) -> None:
        return (text('<!DOCTYPE html>', escape=False, use_parent=False) % CTX).render_now(self)


class TextWrapper(ABCElement):
    __slots__ = (
        'text',
    )

    REPR_TEMPLATE: ClassVar[Callable[..., str]] = '<"{}"{} @{}>'.format

    text: str

    def __init__(
        self,
        text: str,
        /,
        *,
        parent: 'NormalElement[Any] | None' = None,
        escape: bool = True,
        use_parent: bool = MISSING,
        ctx: 'ABCRenderContext | None' = None,
    ) -> None:
        assert \
            (parent is not None) ^ (use_parent is not MISSING), \
            'Should be used only [parent] xor [use_parent].'

        if escape:
            text = html.escape(text)

        self.text = text
        self.parent = parent
        self._ctx = ctx

        if use_parent is not MISSING and use_parent:
            self()

    def __eq__(self: Self, other: ABCElement | object) -> bool:
        return (
            isinstance(other, TextWrapper)
            and self.text == other.text
        )

    def __copy__(self: Self) -> Self:
        copied_self = type(self)(self.text, use_parent=False)
        return copied_self

    def stream(
        self: Self,
        *elems: ABCElement,
        ctx: 'ABCRenderContext' = MISSING,
    ) -> Iterator[str]:
        if self._ctx is not None:
            ctx = self._ctx
        if ctx is MISSING:
            raise RuntimeError('ctx is missing')

        result = ctx.text_element_template(self.text)
        yield ctx.finalizer(ctx, self, result)

        if elems:
            for elem in elems:
                yield from elem.stream(ctx=ctx)

    def __repr__(self) -> str:
        return self.REPR_TEMPLATE(
            self.text[:50],
            '[...]' * (len(self.text) > 50),
            id(self),
        )


class TagElement[TAG: ANY_TAGS](ABCElement):
    __slots__ = (
        'tag',
        'attrs',
    )

    tag: TAG
    attrs: 'Attrs'

    def __eq__(self: Self, other: ABCElement | object) -> bool:
        return (
            isinstance(other, type(self))
            and self.tag == other.tag
            and self.attrs == other.attrs
        )

    def __call__(
        self: Self,
        **attrs: str,
    ) -> Self:
        result = super().__call__()
        if attrs:
            self.attrs.merge_update(attrs)
        return result

    def __lt__(
        self: Self,
        others: dict[str, str] | Sequence[dict[str, str]],
    ) -> Self:
        if isinstance(others, dict):
            others = (others,)
        self.attrs.merge_update(*others)
        return self


class VoidElement[TAG: VOID_TAGS](TagElement[TAG]):
    __slots__ = ()

    REPR_TEMPLATE: ClassVar[Callable[..., str]] = '<{} @{}>'.format

    def __init__(
        self,
        tag: TAG,
        /,
        **attrs: str,
    ) -> None:
        self.tag = tag
        self.attrs = Attrs(attrs)
        self.parent = None
        self._ctx = None
        self()

    @isolate()
    def __copy__(self: Self) -> Self:
        copied_self = type(self)(self.tag, **self.attrs) % self._ctx
        return copied_self

    def stream(
        self: Self,
        *elems: ABCElement,
        ctx: 'ABCRenderContext' = MISSING,
    ) -> Iterator[str]:
        if self._ctx is not None:
            ctx = self._ctx
        if ctx is MISSING:
            raise RuntimeError('ctx is missing')

        tag_content = ctx.element_attrs_template(
            ctx,
            self.tag,
            self.attrs,
        )
        result = ctx.void_element_template(tag_content)

        yield ctx.finalizer(ctx, self, result)

        if elems:
            for elem in elems:
                yield from elem.stream(ctx=ctx)

    def __repr__(self: Self) -> str:
        ctx = self._ctx or CTX

        result = self.REPR_TEMPLATE(
            ctx.element_attrs_template(
                ctx,
                self.tag,
                self.attrs,
            ),
            id(self),
        )

        return result


class NormalElement[TAG: NORMAL_TAGS](TagElement[TAG]):
    __slots__ = (
        '_children',
        '_context_token',
    )

    REPR_TEMPLATE: ClassVar[Callable[..., str]] = '<{} @{}>:{}</{}>'.format

    _children: MutableSequence['ABCElement']
    _context_token: Token | None

    def __init__(
        self,
        tag: TAG,
        /,
        *children: 'ABCElement | str',
        **attrs: str,
    ) -> None:
        self.tag = tag
        self.attrs = Attrs(attrs)
        self.parent = None
        self._ctx = None
        self()
        self._children = list['DOMElement']()
        if children:
            self.add(*children)
        self._context_token = None

    def add(
        self: Self,
        *children: 'ABCElement | str',
    ) -> None:
        new_children: list[ABCElement] = list[Any](children)

        for idx, child in enumerate(new_children):
            if isinstance(child, str):
                new_children[idx] = child = TextWrapper(child, parent=self)
            elif child.parent is not None:
                child.parent.remove(child)
            child.parent = self

        self._children.extend(new_children)

    def insert(
        self: Self,
        index: int,
        *children: 'ABCElement | str',
    ) -> None:
        assert index >= 0
        assert index <= len(self._children)

        new_children: list[ABCElement] = list[Any](children)

        for idx, child in enumerate(new_children):
            if isinstance(child, str):
                new_children[idx] = child = TextWrapper(child, parent=self)
            if child.parent is not None:
                child.parent.remove(child)
            child.parent = self

        moved_children = self._children[index:]
        self._children[index:] = [*new_children, *moved_children]

    def remove(
        self: Self,
        *children: 'ABCElement',
    ) -> None:
        assert all(child in self._children for child in children)

        for child in children:
            self._children.remove(child)
            child.parent = None

    def clear(self: Self) -> None:
        for child in self._children:
            child.parent = None
        self._children.clear()

    @contextmanager
    def overwrite(self: Self) -> Generator[Super, None, None]:
        super_ = Super(*self._children)
        self.clear()

        with self:
            yield super_

    def __eq__(self: Self, other: ABCElement | object) -> bool:
        return (
            isinstance(other, type(self))
            and self.tag == other.tag
            and self.attrs == other.attrs
            and (
                isinstance(other, NormalElement)
                and self._children == other._children
            )
        )

    def __copy__(self: Self) -> Self:
        copied_self = type(self)(self.tag, **self.attrs) % self._ctx

        if self._children:
            for child in self._children:
                copied_child = child.__copy__()
                copied_self._children.append(copied_child)
                copied_child.parent = copied_self

        return copied_self

    def __getstate__(self: Self, *args, **kwargs):
        *results, state = super().__getstate__(*args, **kwargs)  # type: ignore
        state.pop('_context_token', None)
        return *results, state

    def __enter__(self: Self) -> Self:
        self._context_token = _CURRENT_ELEMENT.set(self)
        return self

    def __exit__(
        self: Self,
        exc_type: Optional[type[Exception]],
        exc_value: Optional[Exception],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        assert \
            self._context_token is not None, \
            f'invalid state of {NormalElement.__name__}._context_token'
        _CURRENT_ELEMENT.reset(self._context_token)
        self._context_token = None

    def __iter__(self: Self) -> Iterator[ABCElement]:
        yield self
        for child in self._children:
            yield from child

    def __len__(self: Self) -> int:
        return self._children.__len__()

    def stream(
        self: Self,
        *elems: ABCElement,
        ctx: 'ABCRenderContext' = MISSING,
    ) -> Iterator[str]:
        if self._ctx is not None:
            ctx = self._ctx
        if ctx is MISSING:
            raise RuntimeError('ctx is missing')

        tag_content = ctx.element_attrs_template(
            ctx,
            self.tag,
            self.attrs,
        )

        tag_begin = ctx.normal_element_begin_template(
            tag_content,
        )
        yield ctx.finalizer(ctx, self, tag_begin)

        if self._children:
            for child in self._children:
                yield from child.stream(ctx=ctx)
        tag_end = ctx.normal_element_end_template(self.tag)
        yield tag_end

        if elems:
            for elem in elems:
                yield from elem.stream(ctx=ctx)


    def __repr__(self: Self) -> str:
        ctx = self._ctx or CTX

        result = self.REPR_TEMPLATE(
            ctx.element_attrs_template(
                ctx,
                self.tag,
                self.attrs,
            ),
            id(self),
            len(self._children),
            self.tag,
        )

        return result


text = partial(TextWrapper, use_parent=True)


from sodom.attrs import Attrs
from sodom.contexts import CTX, ABCRenderContext



__license__ = '''
sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''


from operator import itemgetter

from sodom.contexts import PYCTX


if __name__ == '__main__':
    import warnings

    try:
        from lxml import etree as ETree
        from xml.etree.ElementTree import Element, ElementTree
        from typer import Typer
        from loguru import logger
    except ImportError as e:
        warnings.warn('Please install modules required for cli usage `python -m pip install sodom[cli]`')
        raise e
    else:
        from typing import cast
        from pathlib import Path
        from typing import cast

        from sodom import *
        from sodom.elements import NormalElement, VoidElement, ABCElement
        from sodom.guardians import is_normal_element
        from sodom.literals import NORMAL_TAG_STRS, VOID_TAG_STRS, NORMAL_TAGS, VOID_TAGS
        from sodom.contexts import pyctx

        cli = Typer()


        def _build_sodom_document(tree: ElementTree, include_empty_rows: bool) -> NormalElement:

            def build_sodom_element(parent: NormalElement | None, xml_elem: Element) -> ABCElement:
                html_tag = xml_elem.tag
                html_attrs = xml_elem.attrib
                text = (xml_elem.text or '').split('\n')
                text = (
                    t.strip()
                    for t in text
                )
                text = (
                    t
                    for t in text
                    if bool(t) or include_empty_rows
                )

                if html_tag in NORMAL_TAG_STRS:
                    sodom_elem = NormalElement(cast(NORMAL_TAGS, html_tag), *text, **html_attrs)
                elif html_tag in VOID_TAG_STRS:
                    sodom_elem = VoidElement(cast(VOID_TAGS, html_tag), **html_attrs)
                else:
                    raise RuntimeError(f'Unknown tag: {html_tag}')

                if parent is not None:
                    parent.add(sodom_elem)

                for xml_child in xml_elem:
                    build_sodom_element(sodom_elem, xml_child)

                return sodom_elem

            xml_root = tree.getroot()
            sodom_root = build_sodom_element(None, xml_root)

            return sodom_root


        @cli.command(
            help='test',
            short_help='Generate python code via HTML (.html) file.',
        )
        def generator(source: Path, target: Path, dialect: str = 'sodom', include_empty_rows: bool = False):
            logger.info('Parsing source: {}', source)
            tree: ElementTree = ETree.parse(
                source,
                ETree.HTMLParser(
                    remove_comments=True,
                    remove_blank_text=True,
                    collect_ids=False,
                ),
            )

            if dialect == 'sodom':
                logger.info('Building python structure via {}...', 'sodom')
                document = _build_sodom_document(tree, include_empty_rows)
                logger.info('Generating python code via {}...', 'sodom')
                code = document.render(ctx=PYCTX)
            else:
                raise ValueError(f'dialect must be {['sodom']}')

            logger.info('Writing {} target: {}', document, target)
            with open(target, 'w+') as f:
                f.write(code)
            for row in code.split('\n'):
                logger.info(row)
            logger.info('Done!')

        cli()



__license__ = '''
sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import asyncio
from contextlib import nullcontext
import pickle
from concurrent.futures import ThreadPoolExecutor
from random import randbytes
from statistics import mean, median
from time import sleep, time as time_
from typing import Any, Callable, Self, Sequence
import unittest.mock

import pytest

import sodom
from sodom import *
from sodom import literals
from sodom.attrs import Attrs, replace, remove
from sodom.contexts import DEFAULT_HTML_CONTEXT, ABCRenderContext
from sodom.elements import ABCElement, NormalElement, TagElement, VoidElement, text, TextWrapper


def _rand_id():
    return Attrs(id=randbytes(8).hex())


class TestTextElement:
    @pytest.mark.parametrize(
        ('etalon', 'other', 'is_equal'),
        [
            (text('foo'), text('foo'), True),
            (text('foo'), text('bar'), False),
            (text('foo'), object(), False),
        ],
    )
    def test_eq(self: Self, etalon: TextWrapper, other: Any, is_equal: bool):
        assert (etalon == other) == is_equal

    @pytest.mark.parametrize(
        ('parent', 'elem'),
        [
            (nullcontext(), text('foo')),
            (div(), text('foo')),
        ]
    )
    def test_copy(self: Self, parent: NormalElement, elem: TextWrapper):
        with parent:
            copy = elem.__copy__()
        assert all((
            elem.text == copy.text,
            elem.parent is None,
            elem._ctx is None,
        ))


    # @pytest.mark.parametrize(
    #     ('elem', 'result'),
    #     [
    #         (TextWrapper('foo'), 'foo'),
    #         (text('foo'), 'foo'),
    #     ]
    # )
    # def test_stream(self: Self, elem: TextWrapper, result: str):
    #     assert ''.join(elem.stream(ctx=DEFAULT_HTML_CONTEXT)) == result

    @unittest.mock.patch('sodom.contexts.DEFAULT_HTML_CONTEXT')
    def test_stream_uses_context(self, DEFAULT_HTML_CONTEXT: unittest.mock.MagicMock | ABCRenderContext):
        t = text('foo')
        next(iter(t.stream(ctx=DEFAULT_HTML_CONTEXT)))
        DEFAULT_HTML_CONTEXT.text_element_template.assert_called_once_with('foo')
        pass



class TestVoidElement:
    @pytest.mark.parametrize(
        'elem',
        [
            hr(),
            hr(foo='bar'),
        ],
    )
    def test_copy(self, elem: VoidElement):
        copied_elem = elem.__copy__()
        assert elem is not copied_elem
        assert elem == copied_elem

    @pytest.mark.parametrize(
        ('elem', 'rendered'),
        [
            (hr().render_html(), '<!DOCTYPE html><hr>'),
            (hr(foo="bar").render_html(), '<!DOCTYPE html><hr foo="bar">'),
            (hr(foo="bar", baz="").render_html(), '<!DOCTYPE html><hr foo="bar" baz>'),
            (hr(foo="", bar="", baz="").render_html(), '<!DOCTYPE html><hr foo bar baz>'),
        ],
    )
    def test_render(self, elem, rendered):
        assert elem == rendered

    @pytest.mark.parametrize(
        ('elem', 'repred'),
        [
            (repr(elem := hr()), f'<hr @{id(elem)}>'),
            (repr(elem := hr(foo="bar")), f'<hr foo="bar" @{id(elem)}>'),
            (repr(elem := hr(foo="bar", baz="")), f'<hr foo="bar" baz @{id(elem)}>'),
            (repr(elem := hr(foo="", bar="", baz="")), f'<hr foo bar baz @{id(elem)}>'),
        ],
    )
    def test_repr(self, elem, repred):
        assert elem == repred

    def test_parent_changing(self):
        h = hr()
        assert h.parent == None

        with div() as parent1:
            h()
        assert h.parent == parent1
        assert h in parent1._children

        with div() as parent2:
            h()
        assert h.parent == parent2
        assert h in parent2._children


class TestNormalElement:
    @pytest.mark.parametrize(
        'elem',
        [
            div(),
            div(foo='bar'),
            div(hr()),
            div(hr(), foo='bar'),
        ],
    )
    def test_copy(self, elem: NormalElement):
        copied_elem = elem.__copy__()
        assert elem is not copied_elem
        assert elem == copied_elem
        assert elem._children == copied_elem._children

    @pytest.mark.parametrize(
        ('elem', 'rendered'),
        [
            (div().render_html(), '<!DOCTYPE html><div></div>'),
            (div(foo="bar").render_html(), '<!DOCTYPE html><div foo="bar"></div>'),
            (div(foo="bar", baz="").render_html(), '<!DOCTYPE html><div foo="bar" baz></div>'),
            (div(foo="", bar="", baz="").render_html(), '<!DOCTYPE html><div foo bar baz></div>'),

            (div(div()).render_html(), '<!DOCTYPE html><div><div></div></div>'),
            (div(div(), foo="bar").render_html(), '<!DOCTYPE html><div foo="bar"><div></div></div>'),
            (div(div(), foo="bar", baz="").render_html(), '<!DOCTYPE html><div foo="bar" baz><div></div></div>'),
            (div(div(), foo="", bar="", baz="").render_html(), '<!DOCTYPE html><div foo bar baz><div></div></div>'),

            (div(div(), div()).render_html(), '<!DOCTYPE html><div><div></div><div></div></div>'),
            (div(div(), div(), foo="bar").render_html(), '<!DOCTYPE html><div foo="bar"><div></div><div></div></div>'),
            (div(div(), div(), foo="bar", baz="").render_html(), '<!DOCTYPE html><div foo="bar" baz><div></div><div></div></div>'),
            (div(div(), div(), foo="", bar="", baz="").render_html(), '<!DOCTYPE html><div foo bar baz><div></div><div></div></div>'),

            (div(div(div()), div()).render_html(), '<!DOCTYPE html><div><div><div></div></div><div></div></div>'),
            (div(div(div()), div(), foo="bar").render_html(), '<!DOCTYPE html><div foo="bar"><div><div></div></div><div></div></div>'),
            (div(div(div()), div(), foo="bar", baz="").render_html(), '<!DOCTYPE html><div foo="bar" baz><div><div></div></div><div></div></div>'),
            (div(div(div()), div(), foo="", bar="", baz="").render_html(), '<!DOCTYPE html><div foo bar baz><div><div></div></div><div></div></div>'),
        ],
    )
    def test_render(self, elem: str, rendered: str):
        assert elem == rendered

    @pytest.mark.parametrize(
        ('attrs', 'rendered'),
        [
            (Attrs(), ''),
            (Attrs(foo="bar"), ' foo="bar"'),
            (Attrs(foo="bar", baz=""), ' foo="bar" baz'),
            (Attrs(foo="", bar="", baz=""), ' foo bar baz'),
        ],
    )
    def test_render_with_children_via_context(self, attrs, rendered):
        with div(**attrs) as elem:
            pass
        assert elem.render_html() == f'<!DOCTYPE html><div{rendered}></div>'

        with div(**attrs) as elem:
            div()
        assert elem.render_html() == f'<!DOCTYPE html><div{rendered}><div></div></div>'

        with div(**attrs) as elem:
            div()
            div()
        assert elem.render_html() == f'<!DOCTYPE html><div{rendered}><div></div><div></div></div>'

        with div(**attrs) as elem:
            with div():
                div()
            div()
        assert elem.render_html() == f'<!DOCTYPE html><div{rendered}><div><div></div></div><div></div></div>'

    @pytest.mark.parametrize(
        ('elem', 'rendered'),
        [
            (div(), '<div></div>'),
            (div(foo="bar"), '<div foo="bar"></div>'),
            (div(foo="bar", baz=""), '<div foo="bar" baz></div>'),
            (div(foo="", bar="", baz=""), '<div foo bar baz></div>'),

            (div(div()), '<div><div></div></div>'),
            (div(div(), foo="bar"), '<div foo="bar"><div></div></div>'),
            (div(div(), foo="bar", baz=""), '<div foo="bar" baz><div></div></div>'),
            (div(div(), foo="", bar="", baz=""), '<div foo bar baz><div></div></div>'),

            (div(div(), div()), '<div><div></div><div></div></div>'),
            (div(div(), div(), foo="bar"), '<div foo="bar"><div></div><div></div></div>'),
            (div(div(), div(), foo="bar", baz=""), '<div foo="bar" baz><div></div><div></div></div>'),
            (div(div(), div(), foo="", bar="", baz=""), '<div foo bar baz><div></div><div></div></div>'),

            (div(div(div()), div()), '<div><div><div></div></div><div></div></div>'),
            (div(div(div()), div(), foo="bar"), '<div foo="bar"><div><div></div></div><div></div></div>'),
            (div(div(div()), div(), foo="bar", baz=""), '<div foo="bar" baz><div><div></div></div><div></div></div>'),
            (div(div(div()), div(), foo="", bar="", baz=""), '<div foo bar baz><div><div></div></div><div></div></div>'),
        ],
    )
    def test_str(self, elem: NormalElement, rendered: str):
        assert str(elem) == rendered

    @pytest.mark.parametrize(
        ('parent', 'child'),
        [
            (div(), div()),
            (div(div()), div()),
        ],
    )
    def test_add(self, parent: NormalElement, child: NormalElement):
        parent.add(child)
        assert child in parent._children
        assert parent._children[-1] is child

    @pytest.mark.parametrize(
        ('parent', 'child', 'pos'),
        [
            (div(), div(), -1),
            (div(), div(), 0),
            (div(), div(), 1),

            (div(div()), div(), -1),
            (div(div()), div(), 0),
            (div(div()), div(), 1),
            (div(div()), div(), 2),

            (div(div(), div()), div(), -1),
            (div(div(), div()), div(), 0),
            (div(div(), div()), div(), 1),
            (div(div(), div()), div(), 2),
            (div(div(), div()), div(), 3),
        ],
    )
    def test_insert(self, parent: NormalElement, child: NormalElement, pos: int):
        if (0 <= pos <= len(parent._children)):
            ctx = nullcontext()
        else:
            ctx = pytest.raises(AssertionError)

        with ctx:
            parent.insert(pos, child)
            assert parent._children[pos] is child

    @pytest.mark.parametrize(
        ('parent', 'child'),
        [
            (div(div() < _rand_id()), div() < _rand_id()),
            (div((child := div() < _rand_id())), child),

            (div(div() < _rand_id(), div() < _rand_id()), div()),
            (div((child := div() < _rand_id()), div() < _rand_id()), child),
            (div(div() < _rand_id(), (child := div() < _rand_id())), child),

            (div(div() < _rand_id(), div() < _rand_id(), div() < _rand_id()), div()),
            (div((child := div() < _rand_id()), div() < _rand_id(), div() < _rand_id()), child),
            (div(div() < _rand_id(), (child := div() < _rand_id()), div() < _rand_id()), child),
            (div(div() < _rand_id(), div() < _rand_id(), (child := div() < _rand_id())), child),
        ],
    )
    def test_remove(self, parent: NormalElement, child: ABCElement):
        if child in parent._children:
            ctx = nullcontext()
        else:
            ctx = pytest.raises(AssertionError)

        with ctx:
            parent.remove(child)
            assert child not in parent._children
            assert child.parent is None

    def test_pickle(self):
        with div(foo='bar') as doc:
            br()
        p = pickle.dumps(doc)
        unpickled_doc = pickle.loads(p)
        assert doc.render_html(doctype='') == unpickled_doc.render_html(doctype='')

    @pytest.mark.parametrize(
        ('root', 'elems'),
        [
            (
                elem0 := div(
                    elem1 := div(),
                    elem2 := div(
                        elem3 := hr(),
                        elem4 := hr(),
                        elem5 := hr(),
                    ),
                    elem6 := div(),
                    elem7 := div(
                        elem8 := div(),
                        elem9 := div(
                            elem10 := hr(),
                            elem11 := hr(),
                            elem12 := hr(),
                        ),
                        elem13 := div(),
                    ),
                    elem14 := div(),
                ),
                [
                    elem0,
                    elem1,
                    elem2,
                    elem3,
                    elem4,
                    elem5,
                    elem6,
                    elem7,
                    elem8,
                    elem9,
                    elem10,
                    elem11,
                    elem12,
                    elem13,
                    elem14,
                ],
            ),
        ]
    )
    def test_iter(self, root: NormalElement, elems: Sequence[ABCElement]):
        for idx, elem in enumerate(root):
            assert elems[idx] == elem


class TestAttrs:
    @pytest.mark.parametrize(
        ('attrs', 'rendered'),
        [
            ((Attrs(class_='bar'),), 'class="bar"'),
            ((Attrs(_class='bar'),), 'class="bar"'),
            ((Attrs(_class_='bar'),), 'class="bar"'),

            ((Attrs(class__='bar'),), 'class="bar"'),
            ((Attrs(_class__='bar'),), 'class="bar"'),

            ((Attrs(__class='bar'),), 'class="bar"'),
            ((Attrs(__class_='bar'),), 'class="bar"'),

            ((Attrs(__class__='bar'),), 'class="bar"'),
        ],
    )
    def test_stripping_underscores(self, attrs: Sequence[Attrs], rendered: str):
        assert Attrs.join(*attrs).render(ctx=DEFAULT_HTML_CONTEXT) == rendered

    @pytest.mark.parametrize(
        ('attrs', 'rendered'),
        [
            ((Attrs(data_foo='bar'),), 'data-foo="bar"'),
            ((Attrs(data_foo_='bar'),), 'data-foo-="bar"'),
        ],
    )
    def test_replace_underscores(self, attrs: Sequence[Attrs], rendered: str):
        assert Attrs.join(*attrs).render(ctx=DEFAULT_HTML_CONTEXT) == rendered

    @pytest.mark.parametrize(
        ('elem', 'attrs'),
        [
            (hr(), {'foo': 'bar'}),
            (hr(), Attrs(foo='bar')),

            (hr(), ({'foo': 'bar'},)),
            (hr(), (Attrs(foo='bar'),)),
            (hr(), ({'foo': 'bar'}, Attrs(baz='qux'))),

            (hr(), (Attrs(foo='bar'), Attrs(foo='baz'))),

            (div(), {'foo': 'bar'}),
            (div(), Attrs(foo='bar')),

            (div(), ({'foo': 'bar'},)),
            (div(), (Attrs(foo='bar'),)),
            (div(), ({'foo': 'bar'}, Attrs(baz='qux'))),

            (div(), (Attrs(foo='bar'), Attrs(foo='baz'))),

            (hr(foo='bar'), {'foo': 'bar'}),
            (hr(foo='bar'), Attrs(foo='bar')),

            (hr(foo='bar'), ({'foo': 'bar'},)),
            (hr(foo='bar'), (Attrs(foo='bar'),)),
            (hr(foo='bar'), ({'foo': 'bar'}, Attrs(baz='qux'))),

            (hr(foo='bar'), (Attrs(foo='bar'), Attrs(foo='baz'))),

            (div(foo='bar'), {'foo': 'bar'}),
            (div(foo='bar'), Attrs(foo='bar')),

            (div(foo='bar'), ({'foo': 'bar'},)),
            (div(foo='bar'), (Attrs(foo='bar'),)),
            (div(foo='bar'), ({'foo': 'bar'}, Attrs(baz='qux'))),

            (div(foo='bar'), (Attrs(foo='bar'), Attrs(foo='baz'))),
        ],
    )
    def test_lt_attrs(self, elem: TagElement, attrs: dict[str, str] | Attrs | Sequence[dict[str, str] | Attrs]):
        _elem_orig_attrs = elem.attrs.copy()

        elem < attrs

        if isinstance(attrs, Sequence):
            joined_attrs = Attrs.join(_elem_orig_attrs, *attrs)
        else:
            joined_attrs = Attrs.join(_elem_orig_attrs, attrs)

        assert tuple(elem.attrs.items()) == tuple(joined_attrs.items())

    @pytest.mark.parametrize(
        ('elem', 'attrs'),
        [
            (div(), (Attrs(foo='bar'),),),
            (div(), (Attrs(foo='bar'), Attrs(baz='qux')),),
            (div(), (Attrs(foo='bar'), Attrs(foo='baz')),),

            (div(foo='bar'), (Attrs(foo='bar'),),),
            (div(foo='bar'), (Attrs(foo='bar'), Attrs(baz='qux')),),
            (div(foo='bar'), (Attrs(foo='bar'), Attrs(foo='baz')),),
        ],
    )
    def test_call_attr_with_context(self, elem: NormalElement, attrs: Sequence[Attrs]):
        _elem_orig_attrs = elem.attrs.copy()

        with elem:
            for attr in attrs:
                attr()

        joined_attrs = Attrs.join(_elem_orig_attrs, *attrs)
        assert tuple(elem.attrs.items()) == tuple(joined_attrs.items())

    @pytest.mark.parametrize(
        ('attrs', 'rendered'),
        [
            (Attrs(disabled=''), 'disabled'),
        ],
    )
    def test_empty_attr_adding(self, attrs: Attrs, rendered: str):
        assert attrs.render(ctx=DEFAULT_HTML_CONTEXT) == rendered

    @pytest.mark.parametrize(
        ('attrs', 'replacer'),
        [
            (Attrs(disabled=''), Attrs(disabled=replace('disabled'))),
            (Attrs(disabled='true'), Attrs(disabled=replace('disabled'))),
            (Attrs(disabled='true'), Attrs(disabled=replace(''))),
        ],
    )
    def test_replace(self, attrs: Attrs, replacer: Attrs):
        attrs |= replacer
        assert all((
            attrs[k] == w and attrs[k] is w and isinstance(w, replace)
            for k, w in replacer.items()
        ))

    @pytest.mark.parametrize(
        ('attrs', 'remover'),
        [
            (Attrs(disabled=''), Attrs(disabled=remove())),
            (Attrs(disabled='true'), Attrs(disabled=remove())),
            (Attrs(disabled='true'), Attrs(disabled=remove())),
        ],
    )
    def test_remove(self, attrs: Attrs, remover: Attrs):
        attrs |= remover
        assert all((
            k not in attrs and isinstance(w, remove)
            for k, w in remover.items()
        ))

    @pytest.mark.parametrize(
        ('oper', 'left', 'right'),
        [
            (Attrs.merge, Attrs(), Attrs()),
            (Attrs.merge, Attrs(foo='bar'), Attrs()),
            (Attrs.merge, Attrs(), Attrs(foo='bar')),
            (Attrs.merge, Attrs(foo='bar'), Attrs(foo='baz')),
            (Attrs.merge, Attrs(foo='bar', qux='quux'), Attrs(foo='baz')),
            (Attrs.merge, Attrs(foo='bar'), Attrs(foo='baz', qux='quux')),
            (Attrs.merge, Attrs(foo='bar', qux='quux'), Attrs(foo='baz', qux='quux')),

            (Attrs.merge, Attrs(), Attrs(foo='')),
            (Attrs.merge, Attrs(foo=''), Attrs()),
            (Attrs.merge, Attrs(foo=''), Attrs(foo='')),
            (Attrs.merge, Attrs(bar=''), Attrs(foo='')),
            (Attrs.merge, Attrs(foo=''), Attrs(bar='')),
            (Attrs.merge, Attrs(foo='', bar=''), Attrs(foo='', bar='')),

            (Attrs.__or__, Attrs(), Attrs()),
            (Attrs.__or__, Attrs(foo='bar'), Attrs()),
            (Attrs.__or__, Attrs(), Attrs(foo='bar')),
            (Attrs.__or__, Attrs(foo='bar'), Attrs(foo='baz')),
            (Attrs.__or__, Attrs(foo='bar', qux='quux'), Attrs(foo='baz')),
            (Attrs.__or__, Attrs(foo='bar'), Attrs(foo='baz', qux='quux')),
            (Attrs.__or__, Attrs(foo='bar', qux='quux'), Attrs(foo='baz', qux='quux')),

            (Attrs.__or__, Attrs(), Attrs(foo='')),
            (Attrs.__or__, Attrs(foo=''), Attrs()),
            (Attrs.__or__, Attrs(foo=''), Attrs(foo='')),
            (Attrs.__or__, Attrs(bar=''), Attrs(foo='')),
            (Attrs.__or__, Attrs(foo=''), Attrs(bar='')),
            (Attrs.__or__, Attrs(foo='', bar=''), Attrs(foo='', bar='')),
        ],
    )
    def test_merge(self, oper, left: Attrs, right: Attrs):
        new = oper(left, right)
        assert new is not left
        assert new is not right
        assert tuple(new.items()) == tuple(Attrs.join(left, right).items())

    @pytest.mark.parametrize(
        ('oper', 'left', 'right'),
        [
            (Attrs.merge_update, Attrs(), Attrs()),
            (Attrs.merge_update, Attrs(foo='bar'), Attrs()),
            (Attrs.merge_update, Attrs(), Attrs(foo='bar')),
            (Attrs.merge_update, Attrs(foo='bar'), Attrs(foo='baz')),
            (Attrs.merge_update, Attrs(foo='bar', qux='quux'), Attrs(foo='baz')),
            (Attrs.merge_update, Attrs(foo='bar'), Attrs(foo='baz', qux='quux')),
            (Attrs.merge_update, Attrs(foo='bar', qux='quux'), Attrs(foo='baz', qux='quux')),

            (Attrs.merge_update, Attrs(), Attrs(foo='')),
            (Attrs.merge_update, Attrs(foo=''), Attrs()),
            (Attrs.merge_update, Attrs(foo=''), Attrs(foo='')),
            (Attrs.merge_update, Attrs(bar=''), Attrs(foo='')),
            (Attrs.merge_update, Attrs(foo=''), Attrs(bar='')),
            (Attrs.merge_update, Attrs(foo='', bar=''), Attrs(foo='', bar='')),

            (Attrs.__ior__, Attrs(), Attrs()),
            (Attrs.__ior__, Attrs(foo='bar'), Attrs()),
            (Attrs.__ior__, Attrs(), Attrs(foo='bar')),
            (Attrs.__ior__, Attrs(foo='bar'), Attrs(foo='baz')),
            (Attrs.__ior__, Attrs(foo='bar', qux='quux'), Attrs(foo='baz')),
            (Attrs.__ior__, Attrs(foo='bar'), Attrs(foo='baz', qux='quux')),
            (Attrs.__ior__, Attrs(foo='bar', qux='quux'), Attrs(foo='baz', qux='quux')),

            (Attrs.__ior__, Attrs(), Attrs(foo='')),
            (Attrs.__ior__, Attrs(foo=''), Attrs()),
            (Attrs.__ior__, Attrs(foo=''), Attrs(foo='')),
            (Attrs.__ior__, Attrs(bar=''), Attrs(foo='')),
            (Attrs.__ior__, Attrs(foo=''), Attrs(bar='')),
            (Attrs.__ior__, Attrs(foo='', bar=''), Attrs(foo='', bar='')),
        ],
    )
    def test_merge_update(self, oper, left: Attrs, right: Attrs):
        new = oper(left, right)
        assert new is left
        assert new is not right
        assert all((w in left[k] for k, w in right.items()))

    @pytest.mark.parametrize(
        ('attrs'),
        [
            Attrs(class_='foo') | {'class': 'bar'},
            Attrs({'class': 'foo'}) | {'class_': 'bar'},
            Attrs({'class_': 'foo'}) | {'class_': 'bar'},
        ],
    )
    def test_initial_and_merged_attr_name_unmatch(self, attrs: Attrs):
        assert attrs['class'] == 'foo bar'


@pytest.mark.asyncio
async def test_building_html_in_two_tasks():
    for _ in range(10):
        async def task1():
            await asyncio.sleep(2)
            with div() as d:
                await asyncio.sleep(2)
                text('task1')
                text('task1')
            return d

        async def task2():
            with div() as d:
                await asyncio.sleep(2)
                text('task2')
                text('task2')
                await asyncio.sleep(2)
            return d

        div1, div2 = await asyncio.gather(
            task1(),
            task2(),
        )

        assert div1.parent == None
        assert div2.parent == None
        assert div1._children[0] == text('task1')
        assert div1._children[1] == text('task1')
        assert div2._children[0] == text('task2')
        assert div2._children[1] == text('task2')


def test_building_html_in_two_threads():
    for _ in range(10):
        def task1():
            sleep(2)
            with div() as d:
                sleep(2)
                text('task1')
                text('task1')
            return d

        def task2():
            with div() as d:
                sleep(2)
                text('task2')
                text('task2')
                sleep(2)
            return d

        with ThreadPoolExecutor(2) as pool:
            t1 = pool.submit(task1)
            t2 = pool.submit(task2)
        div1 = t1.result()
        div2 = t2.result()

        assert div1.parent == None
        assert div2.parent == None
        assert div1._children[0] == text('task1')
        assert div1._children[1] == text('task1')
        assert div2._children[0] == text('task2')
        assert div2._children[1] == text('task2')


def _sodom_case():
    with sodom.body() as root:
        with sodom.div(class_='d-flex flex-column flex-md-row align-items-center p-3 px-md-4 mb-3 bg-white border-bottom box-shadow'):
            with sodom.h5(class_='my-0 mr-md-auto font-weight-normal'):
                sodom.text('Company name')
            with sodom.nav(class_='my-2 my-md-0 mr-md-3'):
                with sodom.a(class_='p-2 text-dark', href='#'):
                    sodom.text('Features')
                with sodom.a(class_='p-2 text-dark', href='#'):
                    sodom.text('Enterprise')
                with sodom.a(class_='p-2 text-dark', href='#'):
                    sodom.text('Support')
                with sodom.a(class_='p-2 text-dark', href='#'):
                    sodom.text('Pricing')
    result = root.render_html(doctype='')
    return result


def _dominate_case():
    with dominate_tags.body() as root: # type: ignore
        with dominate_tags.div(cls='d-flex flex-column flex-md-row align-items-center p-3 px-md-4 mb-3 bg-white border-bottom box-shadow'): # type: ignore
            with dominate_tags.h5(cls='my-0 mr-md-auto font-weight-normal'): # type: ignore
                dominate_text('Company name')
            with dominate_tags.nav(cls='my-2 my-md-0 mr-md-3'): # type: ignore
                with dominate_tags.a(cls='p-2 text-dark', href='#'): # type: ignore
                    dominate_text('Features')
                with dominate_tags.a(cls='p-2 text-dark', href='#'): # type: ignore
                    dominate_text('Enterprise')
                with dominate_tags.a(cls='p-2 text-dark', href='#'): # type: ignore
                    dominate_text('Support')
                with dominate_tags.a(cls='p-2 text-dark', href='#'): # type: ignore
                    dominate_text('Pricing')
    result = root.render(pretty=False)
    return result


try:
    from dominate import tags as dominate_tags
    from dominate.util import text as dominate_text
except ImportError:
    pass
else:
    def _time_list(func: Callable, runs: int):
        case_time_list = []

        result = None
        for _ in range(runs):
            start = time_()
            result = func()
            current = time_() - start
            case_time_list.append(current)

        sum_sodom_time_list = sum(case_time_list)
        mean_sodom_time_list = mean(case_time_list)
        median_sodom_time_list = median(case_time_list)

        print()
        print(f'\t{func.__name__} :::')
        if result:
            print(f'\t>>> { result.replace('\n', '\n\t>>> ')}')
        print(f'\ttotal={sum_sodom_time_list}; median={median_sodom_time_list}; average={mean_sodom_time_list}')

        return result, (sum_sodom_time_list, median_sodom_time_list, mean_sodom_time_list), case_time_list

    def test_performance_dominate():
        RUNS = 10_000

        sodom_result, (sum_sodom_time_list, median_sodom_time_list, mean_sodom_time_list), _ = _time_list(_sodom_case, RUNS)
        dominate_result, (sum_dominate_time_list, median_dominate_time_list, mean_dominate_time_list), _ = _time_list(_dominate_case, RUNS)

        print()
        ratio_sum = sum_dominate_time_list / sum_sodom_time_list
        ratio_median = median_dominate_time_list / median_sodom_time_list
        ratio_average = mean_dominate_time_list / mean_sodom_time_list
        print(f'\truns: {RUNS}; ratio:: ratio_total={ratio_sum}, ratio_median={ratio_median}, ratio_average={ratio_average}')

        assert sodom_result == dominate_result



__license__ = '''
sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''


from abc import ABC
from dataclasses import dataclass, field
from functools import reduce
from keyword import iskeyword
from operator import truth
import re
from typing import Any, Callable, ClassVar, assert_never
from sodom.guardians import is_normal_element, is_void_element
from sodom.literals import ANY_TAGS


__all__ = [
    'ABCRenderContext',
    'ctx',
    'xctx',
    'pyctx',
    'DEFAULT_HTML_CONTEXT',
    'DEFAULT_XHTML_CONTEXT',
    'DEFAULT_PYTHON_CONTEXT',
    'CTX',
    'XCTX',
    'PYCTX',
]


def _opening_html_tag_content(
    ctx: 'ABCRenderContext',
    tag: ANY_TAGS,
    attrs: 'Attrs',
    /,
) -> str:
    result = ctx.separator.join(
        filter(
            truth,
            (
                tag,
                ctx.separator.join(
                    attrs.stream(ctx=ctx)
                ),
            ),
        ),
    )

    return result


def _opening_python_tag_content(
    ctx: 'ABCRenderContext',
    tag: ANY_TAGS,
    attrs: 'Attrs',
    /,
) -> str:
    result = ''.join(
        filter(
            truth,
            (
                tag,
                '(',
                ctx.separator.join(
                    attrs.stream(ctx=ctx)
                ),
            ),
        ),
    )

    return result


@dataclass(frozen=True, slots=True, kw_only=True)
class ABCRenderContext(ABC):
    quotes: str
    underscore_replacer: str
    separator: str

    text_element_template: Callable[[str], str]
    void_element_template: Callable[[str], str]
    normal_element_begin_template: Callable[[str], str]
    normal_element_end_template: Callable[[str], str]

    attr_template: Callable[..., str]
    element_attrs_template: Callable[['ABCRenderContext', ANY_TAGS, 'Attrs'], str]

    finalizer: Callable[['ABCRenderContext', 'ABCElement', str], str]


@dataclass(frozen=True, slots=True, kw_only=True)
class HTMLRenderContext(ABCRenderContext):
    DEFAULT_QUOTES: ClassVar[str] = '"'
    DEFAULT_UNDERSCORE_REPLACER: ClassVar[str] = '-'
    DEFAULT_SEPARATOR: ClassVar[str] = ' '
    DEFAULT_FINILIZER: ClassVar[Callable[['HTMLRenderContext', 'ABCElement', str], str]] = lambda _, __, s: s

    quotes: str = field(default=DEFAULT_QUOTES)
    underscore_replacer: str = field(default=DEFAULT_UNDERSCORE_REPLACER)
    separator: str = field(default=DEFAULT_SEPARATOR)

    text_element_template: Callable[[str], str] = field(default='{}'.format)
    void_element_template: Callable[[str], str] = field(default='<{}>'.format)
    normal_element_begin_template: Callable[[str], str] = field(default='<{}>'.format)
    normal_element_end_template: Callable[[str], str] = field(default='</{}>'.format)

    attr_template: Callable[..., str] = field(default='{k}={q}{w}{q}'.format)
    element_attrs_template: Callable[['ABCRenderContext', ANY_TAGS, 'Attrs'], str] = \
        field(default=_opening_html_tag_content)

    finalizer: Callable[['HTMLRenderContext', 'ABCElement', str], str] = DEFAULT_FINILIZER


@dataclass(frozen=True, slots=True, kw_only=True)
class PythonRenderContext(ABCRenderContext):
    DEFAULT_QUOTES: ClassVar[str] = '\''
    DEFAULT_UNDERSCORE_REPLACER: ClassVar[str] = '_'
    DEFAULT_SEPARATOR: ClassVar[str] = ', '
    DEFAULT_TABULATION: ClassVar[str] = '\t'

    @staticmethod
    def DEFAULT_FINILIZER(ctx: 'PythonRenderContext', elem: 'ABCElement', s: str) -> str:
        # add tabulation
        parents = tuple(reversed(elem))
        count = len(parents) - 1
        root = parents[-1]
        s = f'{''.join(ctx.tabulation * count)}{s}'
        # add pass to no children element
        if is_normal_element(elem) and not elem._children:
            s = f'{s[:-1]} pass\n'
        # add sodom importing
        if elem.parent is None:
            s = f'from sodom import *\n\n\n{s}'
            s = f'{s[:-2]} as document{s[-2:]}'
        # add root printing
        last = tuple(root)[-1]
        if last is elem:
            s = f'{s}\n\n\ndocument.render_html_now()'
        return s

    @staticmethod
    def DEFAULT_ATTR_REPLATE(*, k: str, q: str, w: str) -> str:
        if iskeyword(k):
            k = f'{k}_'
        k = k.replace('-', '_')
        return '{k}={q}{w}{q}'.format(k=k, q=q, w=w)

    quotes: str = field(default=DEFAULT_QUOTES)
    underscore_replacer: str = field(default=DEFAULT_UNDERSCORE_REPLACER)
    separator: str = field(default=DEFAULT_SEPARATOR)
    tabulation: str = field(default=DEFAULT_TABULATION)

    text_element_template: Callable[[str], str] = field(default='text(\'{}\')\n'.format)
    void_element_template: Callable[[str], str] = field(default='{})\n'.format)
    normal_element_begin_template: Callable[[str], str] = field(default='with {}):\n'.format)
    normal_element_end_template: Callable[[str], str] = field(default=lambda _: '')

    attr_template: Callable[..., str] = field(default=DEFAULT_ATTR_REPLATE)
    element_attrs_template: Callable[['ABCRenderContext', ANY_TAGS, 'Attrs'], str] = \
        field(default=_opening_python_tag_content)

    finalizer: Callable[['PythonRenderContext', 'ABCElement', str], str] = field(default=DEFAULT_FINILIZER)

ctx = HTMLRenderContext
xctx = None  # reserved for XHTML renderer
pyctx = PythonRenderContext

CTX = DEFAULT_HTML_CONTEXT = HTMLRenderContext()
XCTX = DEFAULT_XHTML_CONTEXT = None  # reserved for XHTML renderer
PYCTX = DEFAULT_PYTHON_CONTEXT = PythonRenderContext()


from sodom.elements import ABCElement, NormalElement, VoidElement
from sodom.attrs import Attrs



__license__ = '''
sodom
Copyright (C) 2023, 2024  Dmitry Protasov (inbox@protaz.ru)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License (version 3) as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from functools import reduce
from keyword import iskeyword
from operator import truth
from typing import ClassVar, Iterable, Mapping, Self, final


@final
class remove(str): pass
@final
class replace(str): pass


class Attrs(dict[str, str]):
    DEFAULT_SEPARATOR: ClassVar[str] = ' '

    def __init__(self, *args, **kwargs):
        fake_self = dict(*args, **kwargs)
        super().__init__()

        for k, w in fake_self.items():
            self[k] = w

    def __setitem__(self, attr_name: str, attr_value: str) -> None:
        if iskeyword(stripped_attr_name := attr_name.strip('_')):
            attr_name = stripped_attr_name
        return super().__setitem__(attr_name, attr_value)

    def __call__(self: Self) -> None:
        from sodom.elements import _CURRENT_ELEMENT

        parent = _CURRENT_ELEMENT.get(None)
        assert parent is not None, 'attribute should be called in context of NormalElement'
        parent.attrs.merge_update(self)

    def merge(
        self: Self,
        *others: Mapping[str, str],
        sep: str = DEFAULT_SEPARATOR,
    ) -> 'Attrs':
        '''Merge attributes into new Attrs instance.'''
        copied_self = Attrs(self)
        copied_self.merge_update(*others, sep=sep)
        return copied_self

    def merge_update(
        self: Self,
        *others: Mapping[str, str],
        sep: str = DEFAULT_SEPARATOR,
    ) -> Self:
        '''Merge attributes inplace.'''
        for other in others:
            if other:
                other = Attrs(other)
                for k, v in other.items():
                    if type(v) is remove:
                        del self[k]
                    elif type(v) is replace:
                        self[k] = v
                    else:
                        self[k] = sep.join(filter(
                            truth,
                            (
                                self.get(k, ''),
                                v,
                            ),
                        ))
        return self

    @classmethod
    def join(
        cls: type[Self],
        *others: Mapping[str, str],
        sep: str = DEFAULT_SEPARATOR,
    ) -> 'Attrs':
        result = reduce(lambda l, r: l.merge_update(r, sep=sep), others, cls())
        return result

    def stream(self: Self, *, ctx: 'ABCRenderContext') -> Iterable[str]:
        it = iter(self.items())

        while (attr := next(it, None)) is not None:
            attr_name, attr_value = attr

            if iskeyword(stripped_attr_name := attr_name.strip('_')):
                attr_name = stripped_attr_name

            if ctx.underscore_replacer != '_':
                attr_name = attr_name.replace('_', ctx.underscore_replacer)

            # attribute value must be significant otherwise empty will be rendered.
            if attr_value:
                attr_formatted = ctx.attr_template(k=attr_name, q=ctx.quotes, w=attr_value)
            else:
                attr_formatted = attr_name
            yield attr_formatted

    def render(self: Self, *, ctx: 'ABCRenderContext') -> str:
        it = self.stream(ctx=ctx)
        result = ''.join(it)
        return result

    __or__ = merge
    __ior__ = merge_update  # type: ignore


from sodom.contexts import ABCRenderContext
